

def equal_dicts(d1, d2, ignore_keys):
    ignored = set(ignore_keys)
    for k1, v1 in d1.items():
        if k1 not in ignored and (k1 not in d2 or d2[k1] != v1):
            return False
    for k2, v2 in d2.items():
        if k2 not in ignored and k2 not in d1:
            return False
    return True


def get_coord(string):
    lines = string.split('\n')
    line = len(lines)
    column = len(lines[-1]) + 1
    return line, column
assert(get_coord('aa') == (1, 3))


def get_diff(string):
    lines = string.split('\n')
    line = len(lines) - 1
    column = len(lines[-1])
    return line, column
assert(get_diff('') == (0, 0))
assert(get_diff('aa') == (0, 2))
assert(get_diff('aa\n') == (1, 0))
assert(get_diff('aa\na') == (1, 1))


class BaseType:
    """
    This class is used to count the string-coordinate (line, column) of any element in a statement.
    This is used for identifying, in a script, the line and column of an error.
    It also defines the __eq__
    """
    def __init__(self):
        self._position = None

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return equal_dicts(self.__dict__, other.__dict__, ('_parent', '_parent_index', '_position'))
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    @property
    def position(self):
        if self._position is None:
            raise Exception(self, type(self))
        return self._position

    def set_position(self, position):
        assert(isinstance(position, tuple))
        assert (len(position) == 2)
        self._position = position

    @position.setter
    def position(self, self_position):
        self.set_position(self_position)


class ParserType(BaseType):
    # base type ignored by the interpreter
    pass


class BaseTypeContainer(BaseType):
    """
    This is the base class for containers (e.g. statements, code).

    Relevant of this class:
        * `base_tokens` to get tokens that have functionality.
        * `string_up_to`: the string representation of this class up to an index.
    """
    def __init__(self, tokens):
        super().__init__()
        for i, s in enumerate(tokens):
            assert(isinstance(s, BaseType))
        self._tokens = tokens

    @staticmethod
    def is_base_token(token):
        raise NotImplementedError

    def _as_str(self, func=str):
        raise NotImplementedError

    def _column_delta(self, place='begin'):
        """
        Returns how much the column advances `place=` "begin" or "middle".
        """
        raise NotImplementedError

    def set_position(self, position):
        self._position = position
        position = (position[0], position[1] + self._column_delta())
        for token in self._tokens:
            token.set_position(position)

            token_delta = get_diff(str(token))

            if token_delta[0] == 0:
                initial_column = position[1]
            else:
                initial_column = 1

            position = (
                position[0] + token_delta[0],
                initial_column + token_delta[1] + self._column_delta('middle'))

    @BaseType.position.setter
    def position(self, position):
        super().set_position(position)

    @property
    def tokens(self):
        return self._tokens

    @property
    def base_tokens(self):
        return [token for token in self._tokens if self.is_base_token(token)]

    def __str__(self):
        return self._as_str()



from sqf.types import Code, String, Number, Array, Type, Variable, Boolean, Nothing, Namespace


class InterpreterType(Type):
    # type that is used by the interpreter (e.g. While type)
    def __init__(self, token):
        assert (isinstance(token, Type))
        super().__init__()
        self.token = token

    @property
    def is_undefined(self):
        return self.token.is_undefined


class PrivateType(InterpreterType):
    """
    A type to store the result of "private _x" as in "private _x = 2"
    """
    def __init__(self, variable):
        assert(isinstance(variable, Variable))
        super().__init__(variable)

    @property
    def variable(self):
        return self.token


class WhileType(InterpreterType):
    def __init__(self, condition):
        assert(isinstance(condition, Code))
        super().__init__(condition)

    @property
    def condition(self):
        return self.token


class ForType(InterpreterType):
    def __init__(self, variable=None, from_=None, to=None, step=None):
        if step is None:
            step = Number(1)
        if variable is None:
            variable = String()
        assert (variable is None or isinstance(variable, String))
        assert (from_ is None or isinstance(from_, Type))
        assert (to is None or isinstance(to, Type))
        assert (isinstance(step, Type))
        super().__init__(variable)
        self.from_ = from_
        self.to = to
        self.step = step

    @property
    def variable(self):
        return self.token

    @property
    def is_undefined(self):
        return self.variable.is_undefined or \
               self.from_ is not None and self.from_.is_undefined or \
               self.to is not None and self.to.is_undefined

    def copy(self, other):
        self.token = other.variable
        self.from_ = other.from_
        self.to = other.to
        self.step = other.step


class ForSpecType(InterpreterType):
    def __init__(self, array):
        assert (isinstance(array, Array))
        super().__init__(array)

    @property
    def array(self):
        return self.token


class SwitchType(InterpreterType):
    def __init__(self, keyword, result):
        super().__init__(result)
        self.keyword = keyword

    @property
    def result(self):
        return self.token


class IfType(InterpreterType):
    def __init__(self, condition=None):
        if condition is None:
            condition = Boolean()
        super().__init__(condition)

    @property
    def condition(self):
        return self.token


class ElseType(InterpreterType):
    def __init__(self, then, else_):
        super().__init__(then)
        assert (isinstance(then, Code))
        assert (isinstance(else_, Code))
        self.else_ = else_

    @property
    def then(self):
        return self.token

    @property
    def is_undefined(self):
        return self.else_.is_undefined or self.then.is_undefined


class TryType(InterpreterType):
    def __init__(self, code):
        assert (isinstance(code, Code))
        super().__init__(code)


class WithType(InterpreterType):
    def __init__(self, namespace):
        assert (isinstance(namespace, Namespace))
        super().__init__(namespace)

    @property
    def namespace(self):
        return self.token



from sqf.base_type import ParserType


class Comment(ParserType):

    def __init__(self, string):
        super().__init__()
        if string.startswith('//'):
            self._line = True
        else:
            assert(string.startswith('/*'))
            self._line = False
        self._string = string

    def __str__(self):
        if self._line:
            return '%s' % self._string
        else:
            return '%s' % self._string

    def __repr__(self):
        return ('C(%s)' % self).replace('\n', ' ')


class Space(ParserType):
    def __str__(self):
        return ' '

    def __repr__(self):
        return '\' \''


class Tab(ParserType):
    def __str__(self):
        return '\t'

    def __repr__(self):
        return '\\t'


class EndOfLine(ParserType):
    def __init__(self, value):
        super().__init__()
        assert(value in ['\n', '\r\n'])
        self.value = value

    def __str__(self):
        return self.value

    def __repr__(self):
        return '<EOL>'


class BrokenEndOfLine(ParserType):
    def __str__(self):
        return '\\\n'

    def __repr__(self):
        return '<\EOL>'



# def _check_define_param(param, statement=None):
#     exceptions = []
#     if type(param) == Variable:
#         add = str(param)
#     elif isinstance(param, Keyword):
#         exceptions.append(SQFWarning(param.position, 'redefining keyword "%s".' % str(param)))
#         add = str(param)
#     else:
#         add = None
#         exceptions.append(SQFParserError(param.position, 'invalid 1st argument'))
#
#     if add is not None and statement is not None:
#         if type(statement) != Statement or not statement.parenthesis:
#             exceptions.append(SQFParserError(statement.position, 'wrong construction'))
#             add = None
#
#     return exceptions, add
#
#
# # pre-process #defines
# if base_tokens and base_tokens[0] == Keyword('#define'):
#     tokens = statement._tokens
#     spaces_positions = [i for i,t in enumerate(statement._tokens) if t == Space()]
#     if len(spaces_positions) not in [1, 2]:
#         self.exception(SQFParserError(base_tokens[0].position, '#define must contain exactly 1 or 2 arguments'))
#     elif len(spaces_positions) == 1:
#         self.defines[str(base_tokens[1])] = Nothing
#     elif len(spaces_positions) == 2:
#         arg1 = tokens[spaces_positions[0]+1:spaces_positions[1]]
#         arg2 = tokens[spaces_positions[1]+1:]
#         if len(arg1) == 1:
#             arg1 = arg1[0]
#             exceptions, add = _check_define_param(arg1)
#             if exceptions:
#                 self.exceptions += exceptions
#             if add is not None:
#                 self.defines[add] = arg2
#         elif len(arg1) == 2:  # of the form #define a(...) b
#             exceptions, add = _check_define_param(arg1[0], arg1[1])
#             if exceptions:
#                 self.exceptions += exceptions
#             if add is not None:
#                 self.defines[add] = arg2
#         else:
#             self.exception(SQFParserError(tokens[0].position, '#define 1st argument is invalid'))
#
#     return Nothing
#
#
# def is_parenthesis_statement(statement):
#     if isinstance(statement, Statement):
#         if statement.parenthesis == '()':
#             return True
#         else:
#             assert(len(statement.base_tokens) > 0)
#             return is_parenthesis_statement(statement.base_tokens[0])
#     return False



from sqf.types import Statement, Code, Nothing, Variable, Array, String, Type, File
from sqf.keywords import Keyword
from sqf.exceptions import SQFParserError
import sqf.namespace


class BaseInterpreter:
    """
    Base Interpreter used by the analyzer and interpreter
    """
    def __init__(self, all_vars=None):
        self._namespaces = {
            'uinamespace': sqf.namespace.Namespace('uinamespace'),
            'parsingnamespace': sqf.namespace.Namespace('parsingnamespace'),
            'missionnamespace': sqf.namespace.Namespace('missionnamespace', all_vars),
            'profilenamespace': sqf.namespace.Namespace('profilenamespace')
        }

        self.current_namespace = self.namespace('missionnamespace')

    def exception(self, exception):
        """
        We can overwrite this method to handle exceptions differently
        """
        raise exception

    def set_global_variable(self, var_name, value):
        assert(isinstance(value, Type))
        self.namespace('missionNamespace').base_scope[var_name] = value

    def namespace(self, name):
        return self._namespaces[name.lower()]

    @property
    def current_scope(self):
        return self.current_namespace.current_scope

    def __getitem__(self, name):
        return self.get_scope(name)[name]

    def __contains__(self, name):
        return name in self.get_scope(name)

    def _add_params(self, token):
        self.add_privates([token[0]])

    def add_params(self, base_token):
        assert (isinstance(base_token, Array))
        for token in base_token:
            if isinstance(token, String):
                if token.value == '':
                    continue
                self.add_privates([token])
            elif isinstance(token, Array):
                if len(token) in (2, 3, 4):
                    self._add_params(token)
                else:
                    self.exception(
                        SQFParserError(base_token.position, '`params` array element must have 2-4 elements'))
            else:
                self.exception(SQFParserError(base_token.position, '`params` array element must be a string or array'))

    def value(self, token, namespace_name=None):
        if isinstance(token, Statement):
            return self.value(self.execute_single(statement=token))
        elif isinstance(token, Variable):
            scope = self.get_scope(token.name, namespace_name)
            assert isinstance(scope[token.name], Type)
            return scope[token.name]
        elif isinstance(token, (Type, Keyword)):
            return token
        else:
            raise NotImplementedError(repr(token))

    def get_variable(self, token):
        if isinstance(token, Statement):
            return self.get_variable(token.base_tokens[0])
        else:
            if not isinstance(token, Variable):
                return Nothing()
            return token

    def execute_token(self, token):
        """
        Given a single token, recursively evaluate it and return its value.
        """
        raise NotImplementedError

    def get_scope(self, name, namespace_name=None):
        if namespace_name is None:
            namespace = self.current_namespace
        else:
            namespace = self.namespace(namespace_name)
        return namespace.get_scope(name)

    def add_privates(self, variables):
        for variable in variables:
            assert(isinstance(variable, String))
            name = variable.value

            if not name.startswith('_'):
                self.exception(SQFParserError(variable.position, 'Cannot make global variable "%s" private (underscore missing?)' % name))
            self.current_scope[name] = Nothing()

    def execute_other(self, statement):
        pass

    def execute_code(self, code, params=None, extra_scope=None, namespace_name='missionnamespace'):
        assert (isinstance(code, Code))

        # store the old namespace
        _previous_namespace = self.current_namespace

        # store the executing namespace
        namespace = self.namespace(namespace_name)
        # change to the executing namespace
        self.current_namespace = namespace

        if params is None:
            params = Nothing()
            params.position = code.position
        if extra_scope is None:
            extra_scope = {}
        extra_scope['_this'] = params
        namespace.add_scope(extra_scope)

        # execute the code
        outcome = Nothing()
        outcome.position = code.position
        for statement in code.base_tokens:
            outcome = self.value(self.execute_single(statement))

        # cleanup
        if not isinstance(code, File):  # so we have access to its scope
            # this has to be the executing namespace because "self.current_namespace" may change
            namespace.del_scope()
        self.current_namespace = _previous_namespace
        return outcome

    def execute_single(self, statement):
        """
        Executes a statement
        """
        raise NotImplementedError



import sqf.base_type
from sqf.base_tokenizer import tokenize

from sqf.exceptions import SQFParenthesisError, SQFParserError
from sqf.types import Statement, Code, Number, Boolean, Variable, Array, String, Keyword, Namespace, Preprocessor
from sqf.keywords import KEYWORDS, NAMESPACES, PREPROCESSORS
from sqf.parser_types import Comment, Space, Tab, EndOfLine, BrokenEndOfLine
from sqf.parser_exp import parse_exp


def get_coord(tokens):
    return sqf.base_type.get_coord(''.join([str(x) for x in tokens]))


def identify_token(token):
    """
    The function that converts a token from tokenize to a BaseType.
    """
    if isinstance(token, (Comment, String)):
        return token
    if token == ' ':
        return Space()
    if token == '\t':
        return Tab()
    if token == '\\\n':
        return BrokenEndOfLine()
    if token in ('\n', '\r\n'):
        return EndOfLine(token)
    if token in ('true', 'false'):
        return Boolean(token == 'true')
    try:
        return Number(int(token))
    except ValueError:
        pass
    try:
        return Number(float(token))
    except ValueError:
        pass
    if token in PREPROCESSORS:
        return Preprocessor(token)
    if token.lower() in NAMESPACES:
        return Namespace(token)
    elif token.lower() in KEYWORDS:
        return Keyword(token)
    else:
        return Variable(token)


def parse_strings_and_comments(all_tokens):
    """
    Function that parses the strings of a script, transforming them into `String`.
    """
    string = ''  # the buffer for the activated mode
    tokens = []  # the final result
    in_double = False
    mode = None  # [None, "string_single", "string_double", "comment_line", "comment_bulk"]

    for i, token in enumerate(all_tokens):
        if mode == "string_double":
            string += token
            if token == '"':
                if in_double:
                    in_double = False
                elif not in_double and i != len(all_tokens) - 1 and all_tokens[i+1] == '"':
                    in_double = True
                else:
                    tokens.append(String(string))
                    mode = None
                    in_double = False
        elif mode == "string_single":
            string += token
            if token == "'":
                if in_double:
                    in_double = False
                elif not in_double and i != len(all_tokens) - 1 and all_tokens[i + 1] == "'":
                    in_double = True
                else:
                    tokens.append(String(string))
                    mode = None
                    in_double = False
        elif mode == "comment_bulk":
            string += token
            if token == '*/':
                mode = None
                tokens.append(Comment(string))
                string = ''
        elif mode == "comment_line":
            string += token
            if token == '\n':
                mode = None
                tokens.append(Comment(string))
                string = ''
        else:  # mode is None
            if token == '"':
                string = token
                mode = "string_double"
            elif token == "'":
                string = token
                mode = "string_single"
            elif token == '/*':
                string = token
                mode = "comment_bulk"
            elif token == '//':
                string = token
                mode = "comment_line"
            else:
                tokens.append(token)

    if mode in ("comment_line", "comment_bulk"):
        tokens.append(Comment(string))
    elif mode is not None:
        raise SQFParserError(get_coord(tokens), 'String is not closed')

    return tokens


def _analyze_tokens(tokens):
    ending = ''
    if tokens and tokens[-1] in (Keyword(';'), Keyword(',')):
        ending = tokens[-1].value
        del tokens[-1]

    statement = parse_exp(tokens, container=Statement)
    if isinstance(statement, Statement):
        statement._ending = ending
    else:
        statement = Statement([statement], ending=ending)

    return statement


def _analyze_array_tokens(tokens, tokens_until):
    result = []
    part = []
    first_comma_found = False
    for token in tokens:
        if token == Keyword(','):
            first_comma_found = True
            if not part:
                raise SQFParserError(get_coord(tokens_until), 'Array cannot have an empty element')
            result.append(_analyze_tokens(part))
            part = []
        else:
            part.append(token)

    # an empty array is a valid array
    if part == [] and first_comma_found:
        raise SQFParserError(get_coord(tokens_until), 'Array cannot have an empty element')
    elif tokens:
        result.append(_analyze_tokens(part))
    return result


def parse_block(all_tokens, analyze_tokens, analyze_array, start=0, initial_lvls=None, stop_statement='both'):
    if not initial_lvls:
        initial_lvls = {'[]': 0, '()': 0, '{}': 0}
        initial_lvls.update({x: 0 for x in PREPROCESSORS})
    lvls = initial_lvls.copy()

    statements = []
    tokens = []
    i = start

    while i < len(all_tokens):
        token = all_tokens[i]

        if token == Keyword('['):
            lvls['[]'] += 1
            expression, size = parse_block(all_tokens, analyze_tokens, analyze_array, i + 1, lvls, stop_statement='single')
            lvls['[]'] -= 1
            tokens.append(expression)
            i += size + 1
        elif token == Keyword('('):
            lvls['()'] += 1
            expression, size = parse_block(all_tokens, analyze_tokens, analyze_array, i + 1, lvls, stop_statement)
            lvls['()'] -= 1
            tokens.append(expression)
            i += size + 1
        elif token == Keyword('{'):
            lvls['{}'] += 1
            expression, size = parse_block(all_tokens, analyze_tokens, analyze_array, i + 1, lvls, stop_statement)
            lvls['{}'] -= 1
            tokens.append(expression)
            i += size + 1

        elif token == Keyword(']'):
            if lvls['[]'] == 0:
                raise SQFParenthesisError(get_coord(all_tokens[:i]), 'Trying to close right parenthesis without them opened.')

            if statements:
               raise SQFParserError(get_coord(all_tokens[:i]), 'A statement %s cannot be in an array' % Statement(statements))

            return Array(analyze_array(tokens, all_tokens[:i])), i - start
        elif token == Keyword(')'):
            if lvls['()'] == 0:
                raise SQFParenthesisError(get_coord(all_tokens[:i]), 'Trying to close parenthesis without opened parenthesis.')

            if tokens:
                statements.append(analyze_tokens(tokens))

            return Statement(statements, parenthesis=True), i - start
        elif token == Keyword('}'):
            if lvls['{}'] == 0:
                raise SQFParenthesisError(get_coord(all_tokens[:i]), 'Trying to close brackets without opened brackets.')

            if tokens:
                statements.append(analyze_tokens(tokens))

            return Code(statements), i - start
        elif stop_statement == 'both' and token in (Keyword(';'), Keyword(',')) or \
            stop_statement == 'single' and token == Keyword(';'):
            tokens.append(token)
            statements.append(analyze_tokens(tokens))
            tokens = []
        elif isinstance(token, Keyword) and token.value in PREPROCESSORS and lvls[token.value] == 0:
            if tokens:
                # a pre-processor starts a new statement
                statements.append(analyze_tokens(tokens))
                tokens = []

            # repeat the loop for this token.
            lvls[token.value] += 1
            expression, size = parse_block(all_tokens, analyze_tokens, analyze_array, i, lvls, stop_statement)
            lvls[token.value] -= 1

            statements.append(expression)
            i += size - 1
        elif type(token) == EndOfLine and any(lvls[x] != 0 for x in PREPROCESSORS):
            if tokens:
                if tokens[0] == Preprocessor('#define'):
                    statements.append(Statement(tokens))
                else:
                    statements.append(analyze_tokens(tokens))

            return Statement(statements), i - start
        else:
            tokens.append(token)
        i += 1

    for lvl_type in lvls:
        if lvls[lvl_type] != 0 and lvl_type not in PREPROCESSORS:
            raise SQFParenthesisError(get_coord(all_tokens[:start - 1]), 'Parenthesis "%s" not closed' % lvl_type[0])

    if tokens:
        statements.append(analyze_tokens(tokens))

    return Statement(statements), i - start


def parse(script):
    tokens = [identify_token(x) for x in parse_strings_and_comments(tokenize(script))]

    result = parse_block(tokens, _analyze_tokens, _analyze_array_tokens)[0]

    result.set_position((1, 1))

    return result
